// ======================================================================
// 
// DirectDraw.cpp
// jeff grills
//
// copyright 1997 origin systems inc
//
// ======================================================================

#include "first.h"
#include "DirectDraw.h"

#include "Palette.h"
#include "ConfigPlatform.h"

#include <d3d.h>

// ======================================================================

const u_int PAL_TRANSPARENT_COLOR = 0;

// ======================================================================

HWND                 DirectDraw::windowHandle;
HWND                 DirectDraw::drawWindowHandle;
LPDIRECTDRAW2        DirectDraw::ddraw;
LPDIRECTDRAWSURFACE  DirectDraw::frontSurface;
LPDIRECTDRAWSURFACE  DirectDraw::backSurface;
LPDIRECTDRAWSURFACE  DirectDraw::lockSurface;
LPDIRECTDRAWCLIPPER  DirectDraw::clipper;
DWORD                DirectDraw::totalVideoMemory;

int                  DirectDraw::width;
int                  DirectDraw::height;
int                  DirectDraw::depthBits;

bool                 DirectDraw::oneMegVideoCard;
bool                 DirectDraw::bankSwitched;
bool                 DirectDraw::backBufferInSystemRam;

int                  DirectDraw::rMask;
int                  DirectDraw::gMask;
int                  DirectDraw::bMask;
byte                 DirectDraw::rSizeShift;
byte                 DirectDraw::rPlaceShift;
byte                 DirectDraw::gSizeShift;
byte                 DirectDraw::gPlaceShift;
byte                 DirectDraw::bSizeShift;
byte                 DirectDraw::bPlaceShift;

void                *DirectDraw::buffer;
int                  DirectDraw::pitch;
int                  DirectDraw::lockCount;

int                  DirectDraw::wantPages;
int                  DirectDraw::havePages;
int                  DirectDraw::visiblePage;
int                  DirectDraw::alwaysBlit;

DirectDraw::Mode    *DirectDraw::firstMode;
DirectDraw::Mode    *DirectDraw::currentMode;

int                        DirectDraw::globalDriverId;
DirectDraw::Driver        *DirectDraw::firstDriver;
DirectDraw::Driver        *DirectDraw::lastDriver;
const DirectDraw::Driver  *DirectDraw::driver;

bool                 DirectDraw::mfc;
bool                 DirectDraw::windowed;

// ======================================================================

void DirectDraw::enumerate(void)
{
	// get all the available direct draw devices
	if (!firstDriver)
		DirectDrawEnumerate(directDrawEnumerateCallback, NULL);
}

// ----------------------------------------------------------------------

BOOL WINAPI DirectDraw::directDrawEnumerateCallback(GUID FAR * lpGUID, LPSTR lpDriverDescription, LPSTR lpDriverName, LPVOID /* lpContext */)
{
	Driver *driver;

	DPRINT_LOG(("dd::enum '%s', '%s' %p\n", lpDriverName, lpDriverDescription, lpGUID));

	driver = NEW Driver;
	driver->id     = ++globalDriverId;
	driver->name   = strdup(lpDriverName);
	driver->desc   = strdup(lpDriverDescription);
	driver->pGuid  = lpGUID;
	driver->next   = NULL;

	// copy the guid, but it may be NULL
	if (lpGUID)
	{
		driver->pGuid = &driver->guid;
		driver->guid = *lpGUID;
	}
	else
	{
		driver->pGuid = NULL;
		memset(&driver->guid, 0, sizeof(driver->guid));
	}
	
	// append this driver to the list
	if (!firstDriver)
		firstDriver = driver;
	else
		lastDriver->next = driver;
	lastDriver = driver;

	return DDENUMRET_OK;
}

// ----------------------------------------------------------------------

const DirectDraw::Driver *DirectDraw::getDriver(int id)
{
	const Driver *driver;

	for (driver = firstDriver; driver && driver->id != id; driver = driver->next)
		;

	return driver;
}

// ======================================================================

void DirectDraw::install(HWND newWindowHandle, const Driver *newDriver, bool newMFC)
{
	LPDIRECTDRAW   ddraw1;
	HRESULT        result;
	DDCAPS         ddcaps;
	DDSCAPS        ddscaps;
	DDSURFACEDESC  sd;
	DWORD          freeVideoMemory;
	Mode          *mode;

	DEBUG_FATAL(ddraw, ("DD::install twice"));

	// init variables
	width             = 0;
	height            = 0;
	depthBits         = 0;
	windowHandle      = newWindowHandle;
	drawWindowHandle  = windowHandle;
	alwaysBlit        = ConfigPlatform::blit();
	mfc               = newMFC;
	windowed          = mfc || ConfigPlatform::windowed();

	// default to triple buffered mode
	tripleBuffer();

	// create the directdraw 1 interface
	driver = newDriver;
	MESSAGE(true, ("DD::install using '%s' '%s'\n", driver->name, driver->desc));
	result = DirectDrawCreate(driver->pGuid, &ddraw1, NULL);
	// result = DirectDrawCreate((GUID *)DDCREATE_EMULATIONONLY, &ddraw, NULL);
	// result = DirectDrawCreate((GUID *)DDCREATE_HARDWAREONLY, &ddraw, NULL);
	FATAL_HR("DD::install create %d", result);

	// create the directdraw 2 interface
	result = ddraw1->QueryInterface(IID_IDirectDraw2, (LPVOID *)&ddraw);
	FATAL_HR("DD::install queryinterface dd2 %d", result);

	// destroy the directdraw 1 interface.  we only want to use the dd2 interface
	ddraw1->Release();
	ddraw1 = NULL;

	// set a cooperative level
	ddraw->SetCooperativeLevel(NULL, DDSCL_NORMAL);

	// get the amount of video memory
	memset(&ddcaps, 0, sizeof(ddcaps));
	ddcaps.dwSize = sizeof(ddcaps);
	result = ddraw->GetCaps(&ddcaps, NULL);
	FATAL_HR("DD::install getcaps %d", result);

	// determine if the card is bank switched
	if (ddcaps.dwCaps & DDCAPS_BANKSWITCHED)
		bankSwitched = true;

	// another attempt to get the amount of video ram
	ddscaps.dwCaps = DDSCAPS_VIDEOMEMORY;
	result = ddraw->GetAvailableVidMem(&ddscaps, &totalVideoMemory, &freeVideoMemory);
	if (result == DDERR_NODIRECTDRAWHW)
	{
		// deal with ddraw in complete emulation mode
		WARNING(true, ("DD::getVideoMem GetAvailableVidMem() failed - emulation only\n"));
		totalVideoMemory = 0;
		freeVideoMemory = 0;
	}
	else
		FATAL_HR("DD::install getAvail %d", result);

	// get to the primary surface
	memset(&sd, 0, sizeof(sd));
	sd.dwSize  = sizeof(sd);
	sd.dwFlags = DDSD_CAPS;
	sd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	result = ddraw->CreateSurface(&sd, &frontSurface, NULL);
	FATAL_HR("DD::install createPrimary %d", result);

	// get info about the primary surface
	memset(&sd, 0, sizeof(sd));
	sd.dwSize  = sizeof(sd);
	result = frontSurface->GetSurfaceDesc(&sd);
	FATAL_HR("DD::install getSurfaceDesc %d", result);

	// release the front surface
	frontSurface->Release();
	frontSurface = NULL;

	// add the primary surface contents to the amount of total video memory
#if DIRECTDRAW_VERSION >= 0x0500
	totalVideoMemory += (sd.dwLinearSize * sd.dwHeight);
#else
	totalVideoMemory += (sd.lPitch * sd.dwHeight);
#endif

	if (totalVideoMemory < (1024 * 1024))
		oneMegVideoCard = true;

	// get the display modes
	result = ddraw->EnumDisplayModes(0, NULL, 0, enumDisplayModesCallback);
	FATAL_HR("DD::setupWindow enum disp modes %d", result);

	if (DPrintFlag(DP_P_DDRAW))
	{
		DPRINT_LOG(("DD::install (%d=avail %d=total)=1 (%d=avail %d=total)=2 vid mem\n", 
			ddcaps.dwVidMemFree, ddcaps.dwVidMemTotal, freeVideoMemory, totalVideoMemory));

		DPRINT_LOG(("DD::install %x=blt %x=bltQueue %x=bltStretch %x=bltFill\n",
			(ddcaps.dwCaps & DDCAPS_BLT) != 0,
			(ddcaps.dwCaps & DDCAPS_BLTQUEUE) != 0,
			(ddcaps.dwCaps & DDCAPS_BLTSTRETCH) != 0,
			(ddcaps.dwCaps & DDCAPS_BLTCOLORFILL) != 0));

		DPRINT_LOG(("DD::install %x=overlay %x=overlayNoClip %x=vertBlankInt %x=banked\n",
			(ddcaps.dwCaps & DDCAPS_OVERLAY) != 0,
			(ddcaps.dwCaps & DDCAPS_OVERLAYCANTCLIP) != 0,
			(ddcaps.dwCaps & DDCAPS_VBI) != 0,
			(ddcaps.dwCaps & DDCAPS_BANKSWITCHED) != 0));
	
		for (mode = firstMode; mode; mode = mode->next)
			DPRINT_LOG(("DD::install mode %4dx%4dx%-2d 0x%x=f 0x%x=cc 0x%x=r 0x%x=g 0x%x=b 0x%x=a\n", 
				mode->sd->dwWidth,
				mode->sd->dwHeight,
				mode->sd->ddpfPixelFormat.dwRGBBitCount,
				mode->sd->ddpfPixelFormat.dwFlags,
				mode->sd->ddpfPixelFormat.dwFourCC,
				mode->sd->ddpfPixelFormat.dwRBitMask,
				mode->sd->ddpfPixelFormat.dwGBitMask,
				mode->sd->ddpfPixelFormat.dwBBitMask,
  			mode->sd->ddpfPixelFormat.dwRGBAlphaBitMask));
	}

	// set the cooperative level to fullscreen and exclusive
	result = ddraw->SetCooperativeLevel(windowHandle, (ConfigPlatform::allowReboot() ? DDSCL_ALLOWREBOOT : 0) | DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE | DDSCL_NOWINDOWCHANGES);
	FATAL_HR("DD::setupWindow setcoop %d", result);
}

// ----------------------------------------------------------------------

void DirectDraw::remove(void)
{
	Mode *mode;

	if (ddraw)
	{
		while (firstMode)
		{
			mode = firstMode;
			firstMode = firstMode->next;
			delete [] mode->buffer;
			delete mode;
		}
		firstMode = currentMode = NULL;

		if (clipper)
			clipper->Release();
		if (backSurface && havePages == 0)
			backSurface->Release();
		if (frontSurface)
			frontSurface->Release();

		ddraw->RestoreDisplayMode();
		ddraw->Release();

		clipper = NULL;
		frontSurface = NULL;
		backSurface = NULL;
		ddraw = NULL;
	}

	// whack the list of direct draw drivers
	while (firstDriver)
	{
		lastDriver = firstDriver;
		firstDriver = firstDriver->next;
		delete [] lastDriver->desc;
		delete [] lastDriver->name;
		delete lastDriver;
	}
	firstDriver = NULL;
	lastDriver = NULL;
	driver = NULL;
}

// ----------------------------------------------------------------------

HRESULT WINAPI DirectDraw::enumDisplayModesCallback(LPDDSURFACEDESC sd, LPVOID /* context */)
{
	Mode *mode, *back, *front;

	// create the mode entry and copy the surface desc into it
	mode         = NEW Mode();
	mode->buffer = NEW u_int8[sd->dwSize];
	mode->sd     = (DDSURFACEDESC *)(mode->buffer);
	mode->next   = NULL;
	memcpy(mode->sd, sd, sd->dwSize);

	// find where to insert the mode into our available modes list. sort on width, then height, then bit depth
	for (back = NULL, front = firstMode; front; back = front, front = front->next)
	{
		if (front->sd->dwWidth > sd->dwWidth)
			break;
		if (front->sd->dwWidth == sd->dwWidth)
		{
			if (front->sd->dwHeight > sd->dwHeight)
				break;
			if (front->sd->dwHeight == sd->dwHeight)
			{
				if (front->sd->ddpfPixelFormat.dwRGBBitCount > sd->ddpfPixelFormat.dwRGBBitCount)
					break;
			}
		}
	}

	// insert this mode onto the mode linked list
	if (!back)
	{
		mode->next = firstMode;
		firstMode = mode;
	}
	else
	{
		mode->next = front;
		back->next = mode;
	}

	return DDENUMRET_OK;
}

// ----------------------------------------------------------------------

void DirectDraw::getVideoMem(DWORD caps, DWORD *total, DWORD *free)
{
	HRESULT result;
	DDSCAPS ddscaps;

	ddscaps.dwCaps = caps;
	result = ddraw->GetAvailableVidMem(&ddscaps, total, free);

	// deal with ddraw in complete emulation mode
	if (result == DDERR_NODIRECTDRAWHW)
	{
		DPRINT_LOG(("DD::getVideoMem GetAvailableVidMem() failed - emulation only\n"));

		if (total)
			*total = 0;
		if (free)
			*free = 0;
	}
	else
		FATAL_HR("DD::install getAvail %d", result);
}


// ----------------------------------------------------------------------

void DirectDraw::computeShifts(int colorMask, u_int8 *placeShift, u_int8 *sizeShift)
{
	int mask, place, size;
		
	mask = colorMask;
	for (place = 0; (mask & 1) == 0; mask >>= 1, place++)
		;
	for (size = 8; (mask & 1) == 1; mask >>= 1, size--)
		;

	FATAL(mask, ("dd::computeShifts non-contiguous mask 0x%08x", colorMask));

	*placeShift = static_cast<u_int8>(place);
	*sizeShift  = static_cast<u_int8>(size);
}

// ----------------------------------------------------------------------

void DirectDraw::setDisplayMode(int width, int height, int depthBits, bool anyWindowSize)
{
	HRESULT       result;
	DWORD         hz;
	DDSURFACEDESC sd;

	DEBUG_FATAL(!ddraw, ("DD::setDisplayMode !ddraw"));

	// verify that the requested mode exists
	for (currentMode = firstMode;
			 currentMode && (currentMode->sd->dwWidth != (u_int)width || currentMode->sd->dwHeight != (u_int)height || currentMode->sd->ddpfPixelFormat.dwRGBBitCount != (u_int)depthBits);
			 currentMode = currentMode->next)
		;
	FATAL(!anyWindowSize && !currentMode, ("DD::setDisplayMode no such mode %dx%dx%d", width, height, depthBits));

	// set the display mode
	if (!windowed)
	{
		if (currentMode->sd->dwFlags & DDSD_REFRESHRATE)
			hz = currentMode->sd->dwRefreshRate;
		else
			hz = 0;

		DPRINT_P_DDRAW(("DD::setDisplayMode %d %d %d %d\n", width, height, depthBits, (int)hz));

		do
		{
			result = ddraw->SetDisplayMode(width, height, depthBits, 0, 0);
		} while (result == DDERR_WASSTILLDRAWING);
		FATAL_HR("DD::setDisplayMode %d", result);
	}

	// get the current video mode
	memset(&sd, 0, sizeof(sd));
	sd.dwSize  = sizeof(sd);
	result = ddraw->GetDisplayMode(&sd);
	FATAL_HR("DD::setDisplayMode getDM %d", result);

	// verify the BPP of the display
	FATAL(sd.ddpfPixelFormat.dwRGBBitCount != (u_int)depthBits, ("DD::setDisplayMode windowed with wrong bit depthBits %d=want %d=is", depthBits, sd.ddpfPixelFormat.dwRGBBitCount));

	// figure out the color style to be using
	if (sd.ddpfPixelFormat.dwRGBBitCount == 8)
	{
		DPRINT_LOG(("DD::setDisplayMode pal 8\n"));
	}
	else
	{
		rMask = sd.ddpfPixelFormat.dwRBitMask;
		gMask = sd.ddpfPixelFormat.dwGBitMask;
		bMask = sd.ddpfPixelFormat.dwBBitMask;

		computeShifts(rMask, &rPlaceShift, &rSizeShift);
		computeShifts(gMask, &gPlaceShift, &gSizeShift);
		computeShifts(bMask, &bPlaceShift, &bSizeShift);
		DPRINT_LOG(("DD::setDisplayMode %04x,%d,%d=r %04x,%d,%d=g %04x,%d,%d=b\n", rMask, rSizeShift, rPlaceShift, gMask, gSizeShift, gPlaceShift, bMask, bSizeShift, bPlaceShift));
	}
	
	result = ddraw->GetMonitorFrequency(&hz);

	if (result == DD_OK)
		DPRINT_LOG(("DD::setDisplayMode monitor frequency %d=hz*100\n", (int)hz));
	else
		DPRINT_LOG(("DD::setDisplayMode could not get monitor frequency\n"));
}

// ----------------------------------------------------------------------

void DirectDraw::createSurfaces(int _width, int _height, int _depthBits)
{
	HRESULT       result;
	DDSURFACEDESC sd;
	DDSCAPS       ddscaps;

	visiblePage = 0;
	width       = _width;
	height      = _height;
	depthBits   = _depthBits;

	// remove any surfaces we had before
	if (clipper)
		clipper->Release();
	if (backSurface && havePages == 0)
		backSurface->Release();
	if (frontSurface)
		frontSurface->Release();
	clipper = NULL;
	frontSurface = NULL;
	backSurface = NULL;

	backBufferInSystemRam = false;
		
	// change the screen resolution appropriately
	setDisplayMode(width, height, depthBits, ConfigPlatform::anyWindowSize());

	// cannot flip if the back buffer is in system ram or if we're running in a window
	havePages = (ConfigPlatform::backBufferNotInVRAM() || windowed) ? 0 : wantPages;
	
	// attempt to get as many flippable buffers as requested
	if (havePages)
		do
		{
			// attempt to create page flipping surfaces in vram
			memset(&sd, 0, sizeof(sd));
			sd.dwSize  = sizeof(sd);
			sd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
			sd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP | DDSCAPS_VIDEOMEMORY | DDSCAPS_3DDEVICE;
			sd.dwBackBufferCount = havePages;
			result = ddraw->CreateSurface(&sd, &frontSurface, NULL);

			DPRINT_LOG(("DD::createSurfaces attempting %d backbuffers, HR=%d \n", havePages, HRESULT_CODE(result)));
		} while (result != DD_OK && --havePages != 0);

	// see if we created flipping surfaces
	if (havePages)
	{
		// get the attached back surface for our render target
		ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
		result = frontSurface->GetAttachedSurface(&ddscaps, &backSurface);		
		FATAL_HR("DD::createSurfaces getAttachedSurface %d", result);
		DPRINT_LOG(("DD::createSurfaces page flip %d=backbuffers\n", havePages));
	}
	else
	{
		// get the primary surface
		memset(&sd, 0, sizeof(sd));
		sd.dwSize  = sizeof(sd);
		sd.dwFlags = DDSD_CAPS;
		sd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
		result = ddraw->CreateSurface(&sd, &frontSurface, NULL);
		FATAL_HR("DD::createSurfaces primary failed", result);

		// setup the back surface description
		memset(&sd, 0, sizeof(sd));
		sd.dwSize  = sizeof(sd);
		sd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
		sd.dwWidth = width;
		sd.dwHeight = height;

		if (!ConfigPlatform::backBufferNotInVRAM())
		{
			// try to create the back surface in video ram
			sd.ddsCaps.dwCaps = DDSCAPS_3DDEVICE | DDSCAPS_VIDEOMEMORY;
			result = ddraw->CreateSurface(&sd, &backSurface, NULL);
			if (result == DD_OK) 
				DPRINT_LOG(("DD::createSurfaces back buffer seperate in video ram\n"));
			else
				DPRINT_LOG(("DD::createSurfaces back buffer could not be created in video ram\n"));
		}

		if (ConfigPlatform::backBufferNotInVRAM() || result != DD_OK)
		{
			// try to create the back surface in system ram
			sd.ddsCaps.dwCaps = DDSCAPS_3DDEVICE | DDSCAPS_SYSTEMMEMORY;
			result = ddraw->CreateSurface(&sd, &backSurface, NULL);
			FATAL_HR("DD::createSurfaces back buffer in system ram failed", result);
			WARNING(true, ("DD::createSurfaces back buffer in system ram\n"));

			backBufferInSystemRam = true;
		}
	}

	// create and attach the clipper if we operating inside a window on the primary
	if (windowed)
	{
		result = ddraw->CreateClipper(0, &clipper, NULL);
		FATAL_HR("DD::createSurfaces createClipper %d", result);

		result = clipper->SetHWnd(0, drawWindowHandle);
		FATAL_HR("DD::createSurfaces clipper sethwnd %d", result);

		result = frontSurface->SetClipper(clipper);
		FATAL_HR("DD::createSurfaces setclipper %d", result);
	}

	// clear our newly created front and back surfaces
	if (clear(0))
	{
		slam();
		clear(0);
	}
}

// ----------------------------------------------------------------------

void DirectDraw::setDrawWindow(HWND hwnd)
{
	HRESULT result;

	drawWindowHandle = hwnd;
	result = clipper->SetHWnd(0, drawWindowHandle);
	FATAL_HR("DD::createSurfaces clipper sethwnd %d", result);
}

// ----------------------------------------------------------------------

bool DirectDraw::createSurface(LPDIRECTDRAWSURFACE *surf, DDSURFACEDESC *desc, LPDIRECTDRAWPALETTE ddpal, bool allowFail)
{
	HRESULT result;

	DEBUG_FATAL(!ddraw, ("DD::createSurface !ddraw"));

	result = ddraw->CreateSurface(desc, surf, NULL);
	if (allowFail && result == DDERR_OUTOFVIDEOMEMORY)
		return false
	FATAL_HR("DD::createSurface cs %d", result);

	if (ddpal)
	{
		// bind the palette to the surface
		result = (*surf)->SetPalette(ddpal);
		FATAL_HR("DD::createSurface SetPal %d", result);
	}

	return true;
}

// ----------------------------------------------------------------------

void DirectDraw::createPalette(LPDIRECTDRAWPALETTE *ddpal, const RGB *pal, int palEntries)
{
	int                 i;
	HRESULT             result;
	PALETTEENTRY        entries[256];

	// initialize palette entries
	for (i = 0; i < palEntries; i++, pal++)
	{
		entries[i].peRed   = pal->r;
		entries[i].peGreen = pal->g;
		entries[i].peBlue  = pal->b;
		entries[i].peFlags = PC_NOCOLLAPSE | D3DPAL_READONLY;
	}

	// clear the other palette entries to bright ugly pink
	for (; i < 256; i++)
	{
		entries[i].peRed   = 0xff;
		entries[i].peGreen = 0x00;
		entries[i].peBlue  = 0xff;
		entries[i].peFlags = PC_NOCOLLAPSE | D3DPAL_READONLY;
	}

	// create the dd palette
	result = ddraw->CreatePalette(DDPCAPS_8BIT|DDPCAPS_ALLOW256|DDPCAPS_INITIALIZE, entries, ddpal, NULL);
	FATAL_HR("DD::bindPal CreatePal %d",result);
}

// ----------------------------------------------------------------------

int DirectDraw::screenLock(bool backBuffer)
{
	HRESULT result;
	DDSURFACEDESC surfDesc;

	if (backBuffer)
		lockSurface= backSurface;
	else
		lockSurface = frontSurface;

	if (lockCount++ == 0)
	{
		memset(&surfDesc, 0, sizeof(surfDesc));
		surfDesc.dwSize = sizeof(surfDesc);

		// handle front surface getting lost
		result = frontSurface->IsLost();
		if (result == DDERR_SURFACELOST)
			return 0;

		// lock the proper surface
		result = lockSurface->Lock(NULL, &surfDesc, DDLOCK_WAIT, NULL);
		if (result == DDERR_SURFACELOST)
			return 0;
		else
			FATAL_HR("DD::screenLock failed %d", result);

#if DIRECTDRAW_VERSION >= 0x0500
 		pitch  = surfDesc.dwLinearSize;
#else
 		pitch  = surfDesc.lPitch;
#endif

		buffer = surfDesc.lpSurface;
		DEBUG_FATAL(!buffer, ("DD::screenLock NULL buffer"));
	}

	return 1;
}

// ----------------------------------------------------------------------

void DirectDraw::screenUnlock(void)
{
	HRESULT result;

	if (--lockCount == 0)
	{
		DEBUG_FATAL(!buffer, ("DD::screenLock no buffer"));
		result = lockSurface->Unlock(buffer);
		FATAL_HR("DD::screenUnlock failed %d", result);
		
		buffer = NULL;
		pitch = -1;
		lockSurface = NULL;
	}
}

// ----------------------------------------------------------------------

bool DirectDraw::clear(u_int32 background)
{
	HRESULT result;
	DDBLTFX bltFX;

	DEBUG_FATAL(lockCount, ("DD::clear while locked"));
	memset(&bltFX, 0, sizeof(bltFX));
	bltFX.dwSize = sizeof(bltFX);
	bltFX.dwFillColor = background;

	for (;;)
	{
		result = backSurface->Blt(NULL, NULL, NULL, DDBLT_COLORFILL|DDBLT_WAIT, &bltFX);
		switch (result)
		{
			case DD_OK:
				return true;
		
			case DDERR_SURFACELOST:
				// fail and recreate the gll
				return false;

			case DDERR_SURFACEBUSY:
				// try again
				break;

			default:
				// hard failure
				FATAL_HR("DD::clear %d", result);
				return false;
		}
	}
}

// ----------------------------------------------------------------------

void DirectDraw::slam(void)
{
	HRESULT result;
	RECT    src, dest;
	int     x, y, done;

	DEBUG_FATAL(!ddraw, ("DD::slam !ddraw"));

	if (havePages && !alwaysBlit)
	{
		result = frontSurface->Flip(NULL, DDFLIP_WAIT);
		FATAL_HR("DD::slam flip %d", result);
		if (visiblePage++ == havePages)
			visiblePage = 0;
	}
	else
	{
		// copy the whole source
		src.left   = 0;
		src.top    = 0;
		src.right  = width;
		src.bottom = height;

		// copy to the client area of the window
		GetWindowRect(drawWindowHandle, &dest);

		if (mfc)
		{
			x = dest.left + 2;
			y = dest.top + 2;
		}
		else
		{
			x = dest.left + GetSystemMetrics(SM_CXFRAME);
			y = dest.top + GetSystemMetrics(SM_CYFRAME) + GetSystemMetrics(SM_CYCAPTION);
		}

		dest.left   = x;
		dest.top    = y;
		dest.right  = x + width;
		dest.bottom = y + height;

		done = 0;
		while (!done)
		{
			result = frontSurface->Blt(&dest, backSurface, &src, DDBLT_WAIT, NULL);

			switch (result)
			{
				case DD_OK:
					done = 1;
					break;

				case DDERR_SURFACELOST:
					done = 1;
					break;

				default:
					FATAL_HR("DD::slam blt failed %d", result);
			}
		}
	}
}

// ----------------------------------------------------------------------
